﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using CTNet;

namespace ChkBoxCtrl
{
    public partial class Form1 : Form
    {
        public Form1(string[] args)
        {
            InitializeComponent();

            // For want of a better place for this credit: The checkbox icon is
            // compliments of http://icons8.com

            if (editorLaunch == false)
            {
                // Set the Form startup as a stand alone app title.
                this.Text = string.Format("{0} {1}", wndwTitle, version);
            }

            // Setting focus to the 'State Box' area.
            stateBox.Select();
        }

        private string FindDocZipFile(string zipFileName)
        {
            string location = null;
            string msg = "";
            string userZipFileName = "";

            bool getZip = true;

            OpenFileDialog openFileDialog = new System.Windows.Forms.OpenFileDialog();

            openFileDialog.Title = string.Format("Check Box Control needs your help in locating: {0}", zipFileName);
            openFileDialog.InitialDirectory = ChkBoxCtrl.Form1.curDir;
            openFileDialog.FileName = zipFileName;
            openFileDialog.DefaultExt = "zip";
            openFileDialog.Filter = string.Format("|{0}", zipFileName);

            try
            {
                do
                {
                    DialogResult result = openFileDialog.ShowDialog(this);

                    if (result == DialogResult.OK)
                    {
                        userZipFileName = Path.GetFileName(openFileDialog.FileName);

                        if (userZipFileName.ToUpper() == zipFileName.ToUpper())
                        {
                            location = openFileDialog.FileName;

                            getZip = false;
                        }
                        else
                        {
                            openFileDialog.FileName = zipFileName;

                            msg = string.Format("Check Box Control needs to know where this zip file is located:\n\n {0}", zipFileName);
                            MessageBox.Show(msg, "Incorrect File Name:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                        }
                    }
                    // Cancel button was pressed.
                    else if (result == DialogResult.Cancel)
                    {
                        getZip = false;
                    }
                } while (getZip == true);
            }
            catch (Exception ex)
            {
                location = null;

                MessageBox.Show(ex.ToString(), "FindDocZipFile():", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            return location;
        }

        private void helpButton_Click(object sender, EventArgs e)
        {
            //MessageBox.Show(string.Format("Current Form1 Height: {0}", this.Height));
            if (showHelp)
            {
                // No more help for now..and no more soup for you!
                showHelp = false;

                if (setList.SelectedIndices.Count > 0)
                {
                    xmlFilterSetListButton.Enabled = true;
                }
                else
                {
                    xmlFilterSetListButton.Enabled = xmlFiltBtnEnabled;
                }

                helpList.Items.Clear();
                helpListLabel.Text = helpText;
                helpButton.BackColor = Color.LightGray;
                helpButton.ForeColor = Color.Blue;
                helpPanel.Visible = false;
                this.CenterToScreen();
            }
            else
            {
                showHelp = true;

                xmlFiltBtnEnabled = xmlFilterSetListButton.Enabled;
                xmlFilterSetListButton.Enabled = true;

                helpButton.BackColor = Color.Gray;
                helpButton.ForeColor = Color.Yellow;

                helpPanel.Visible = true;
                this.CenterToScreen();
            }
        }

        /*
         * For some strange reason, this event handler and it's this.Visible
         * call is required in order to see the form window
         * upon startup.
         */
        private void Form1_Shown(object sender, EventArgs e)
        {
            this.Visible = true;

            // I found that this will select the form with the cursor
            // in the stateBox upon startup.
            this.Select();

            if (editorLaunch == true)
            {
                string xmlFileName;

                /*
                 * ProcessArgs(string[] args) verifies the file info
                 * when launched from an editor. We can not call
                 * FindDocZipFile() from there else the form will be
                 * parked on the taskBar requiring the user to click
                 * the icon in order to bring up the form. Se we do all
                 * of this stuff here.
                 */
                if (vscFileInfo.verifyResult == verifyFile.noZipFile)
                {
                    vscFileInfo.fileTmp = FindDocZipFile(vscFileInfo.fileNameTmp);

                    if (string.IsNullOrEmpty(vscFileInfo.fileTmp))
                    {
                        MessageBox.Show(string.Format("  The document's zip file can not be located:\n\n  {0}\n\n  Launch will be aborted.", vscFileInfo.fileNameTmp), "Check Box Control: Launch Error", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button1, MessageBoxOptions.DefaultDesktopOnly);

                        Application.Exit();
                    }
                }

                // We started from an editor - ProcessArgs(args) has checked the supplied args.
                modeToolStripMenuItem.Enabled = true;

                // Show the file name and path in the window title bar.
                this.Text = string.Format("{0} {1} - {2}", wndwTitle, version, vscFileInfo.launchFileTmp);

                // Launched from Visual Studio Code editor. Any selected text is in clipboard.
                chkBoxCtrlStr = (Clipboard.ContainsText(TextDataFormat.Text) ? Clipboard.GetText(TextDataFormat.Text) : "");

                /*
                 * Now inspect the contents of chkBoxCtrlStr for:
                 * 
                 * If any checkbox sets we will determine the box format - either
                 * the old format (box1_10) or the new format (box12).
                 * 
                 * Check box sets, if we find any we set launchMode to 'Edit Check
                 * Box Sets' else it will be set to 'Add Check Box Control'.
                 * 
                 * The only thing the user can send [anything pertinent] to this application
                 * are Check Box Sets for editing. Of course they can select 'whatever' but
                 * the conclusion would be they must have launched this tool to add check box
                 * control to a document since check box sets were not detected.
                 */
                ProcessInputCode(true);

                /*
                 * Set the msgLabel to indicate the application will exit after they click
                 * either 'Copy Check Box Sets To Clipboard' button or both 'Copy Method And
                 * Method Call To Clipbosrd' and 'Copy Check Box Sets To Clipboard' buttons.
                 * 
                 * Also set the modeLabel to indicate the current mode.
                 */
                msgLabel.BackColor = Color.Red;
                ProcessMode(launchMode, true);

                /*
                 * Checkbox Control 'imprints' on this file name when
                 * launched from the code editor. The complete path and
                 * full file name is also stored for use by ctClient
                 * when communicating with CalcHelper. We do not alter
                 * this file name.
                 */
                xmlIdLabel.Text = vscFileInfo.fileNameNoExtTmp;

                if (vscFileInfo.foundXmlFile == true)
                {
                    SetXMLCheckBoxList(vscFileInfo.fileTmp);
                }
                else
                {
                    // We know we have a zip file else we would not be here.
                    xmlFileName = UnzipXmlFile(vscFileInfo.fileTmp);

                    if (!String.IsNullOrEmpty(xmlFileName))
                    {
                        //MessageBox.Show(xmlFileName);

                        SetXMLCheckBoxList(xmlFileName);

                        // Delete the .xml file we just extracted.
                        File.Delete(xmlFileName);
                    }
                }

                if (vscFileInfo.portFileFound == true)
                {
                    /*
                     * We need to do this here because we can not call
                     * Application.Exit() from a [critical] ctClient exception
                     * or error condition during ctClient.StartClient() when 
                     * ctClient.StartClient() is called from the Form1 constructor
                     * as we get a disposed object exception since the Form1 will
                     * be attempting to open/accessed while it is already disposed of.
                     */
                    ctClient = new CTClient(vscFileInfo);

                    /*
                     * Subscribe to ctClient's event so we can be informed concerning
                     * communication with the calc tool server as it relates to the
                     * Calc Helper tool.
                     */
                    ctClient.CtClientEvent += ctClient_CtClientEvent;

                    ctClient.StartClient();
                }
                else
                {
                    string errMsg;

                    errMsg = string.Format("  The Calc Tool Server port file:\n\n  {0}\n\n  Could not be found.\n\n  Checkbox Control will not be able to ask\n  CalcHelper for checkbox selections.", vscFileInfo.portFileTmp);

                    MessageBox.Show(errMsg, "Calc Tool Server Port File Could Not Be Found:", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        // Calc Tool Client event handler:
        private void ctClient_CtClientEvent(object sender, CtClientArgs e)
        {
            if (e != null)
            {
                switch (e.eventType)
                {
                    case (ctEvent.boxSel):

                        /*
                         * The box selection message from CalcHelper contains
                         * a front end delimited by the pipe char ('|') which
                         * tells us what type of selection message we have:
                         * 
                         * 'boxSets' - This is after CalcHelper acknowledges our
                         *     request for box selections. The boxes are the
                         *     Checkbox control box sets that CalcHelper keeps
                         *     track of during Refreshing it's controlled box
                         *     status and also done when CalcHelper is launched.
                         *     
                         *     The checkbox sets are automatically filtered to
                         *     remove duplicate boxes and orphan checkboxes prior
                         *     to sending to Checkbox Control.
                         * 
                         * 'delBoxSet' - The box set indicated by the accompanying
                         *      index has been deleted in Calc Helper. We must also
                         *      delete that set here.
                         *      
                         * 'updateSel' - Replace the currently selected box set
                         *      with the box set in this message.
                         * 
                         * 'newCHSelIndex' - Calc Helper has selected a new box set.
                         *      All we have in this message is the new selection index.
                         * 
                         * 'newCHSet' - Calc Helper added a new checkbox set to the end
                         *      of it's set list, we will do the same and select it and
                         *      set our boxSetsIndex accordingly.
                         */

                        string msgType = e.eventMsg.Substring(0, e.eventMsg.IndexOf('|'));
                        string msgData = e.eventMsg.Substring(e.eventMsg.IndexOf('|') + 1);
                        int selIndex;
                        boxSelMsgTypes msgTypeEnum = (boxSelMsgTypes)Enum.Parse(typeof(boxSelMsgTypes), msgType);

                        if (Form.ActiveForm != this)
                        {
                            this.Activate();
                        }

                        if (this.WindowState == FormWindowState.Minimized)
                        {
                            this.WindowState = FormWindowState.Normal;
                        }

                        switch (msgTypeEnum)
                        {
                            case (boxSelMsgTypes.updateSel):

                                ProcessCHBoxSetUpdate(msgData);

                                break;

                            case (boxSelMsgTypes.boxSets):

                                ProcessCHBoxSets(msgData);

                                break;

                            case (boxSelMsgTypes.newCHSet):

                                //MessageBox.Show(string.Format("Calling ProcessNewCHBoxSet():\n\nboxSet = \n\n{0}", msgBoxes));

                                ProcessNewCHBoxSet(msgData);

                                break;

                            case (boxSelMsgTypes.delBoxSet):

                                // We know this will work.
                                int.TryParse(msgData, out selIndex);

                                //MessageBox.Show(string.Format("Calling ProcessNewCHSelIndex():\n\nselIndex = {0}", selIndex));

                                ProcessSetListSetDeletion("CH", selIndex);

                                break;

                            case (boxSelMsgTypes.newCHSelIndex):

                                // We know this will work.
                                int.TryParse(msgData, out selIndex);

                                //MessageBox.Show(string.Format("Calling ProcessNewCHSelIndex():\n\nselIndex = {0}", selIndex));

                                ProcessNewCHSelIndex(selIndex);

                                break;
                        }

                        break;

                    case (ctEvent.docLockOrNoCheckboxes):

                        MessageBox.Show(this, "  CalcHelper has denied your request for checkboxes\n  because the document either:\n\n  A) Has no checkboxes\n  B) Is not connected to the Vscode File", "Either - No Checkboxes or Not Connected To Vscode File:", MessageBoxButtons.OK, MessageBoxIcon.Information);

                        ProcessCHBoxSelOnOff(ctClient.boxSelFromCH);

                        break;

                    case (ctEvent.disconnectCH):

                        /*
                         * The server has told us that CalcHelper has closed. If the user
                         * now clicks to ask for boxes the server will inform the user that
                         * they need to launch CalcHelper.
                         */
                        if (Form.ActiveForm != this)
                        {
                            this.Activate();
                        }

                        if (this.WindowState == FormWindowState.Minimized)
                        {
                            this.WindowState = FormWindowState.Normal;
                        }

                        ProcessCHBoxSelOnOff(ctClient.boxSelFromCH);

                        break;

                    case (ctEvent.boxSelOff):

                        /*
                         * Calc Helper is telling us we will no longer be
                         * getting check boxes. Either Calc Helper is being
                         * shut down, or user has chosen to do this, or CalcHelper
                         * has entered lockdown which happens if the user switches
                         * to a different file in the code editor.
                         */
                        if (Form.ActiveForm != this)
                        {
                            this.Activate();
                        }

                        if (this.WindowState == FormWindowState.Minimized)
                        {
                            this.WindowState = FormWindowState.Normal;
                        }

                        ProcessCHBoxSelOnOff(ctClient.boxSelFromCH);

                        break;

                    case (ctEvent.boxSelAck):

                        /*
                         * Calc Helper has acknowledged our request for
                         * check boxes.
                         */
                        ProcessCHBoxSelOnOff(ctClient.boxSelFromCH);

                        break;

                    case (ctEvent.connected):

                        // Check Box Control has connected to Calc Tool Server.
                        isConnected = true;
                        vsCodeIconLbl.Image = ChkBoxCtrl.Properties.Resources.VscodeGreen;
                        vsCodeIconLbl.ToolTipText = "Connected To Calc Tool Server";
                        boxSelFromCHBtn.Enabled = true;
                        boxSelFromCHBtn.ToolTipText = "Click To Receive Checkbox Selections From CalcHelper";

                        break;

                    case (ctEvent.connectFail):

                        // Unsubscribe from ctClient's event. Backout gracefully.
                        ctClient.CtClientEvent -= ctClient_CtClientEvent;
                        ctClient = null;
                        vsCodeIconLbl.ToolTipText = "Not Connected To Calc Tool Server";

                        break;

                    case (ctEvent.shutdown):

                        /* 
                         * Code editor is going down, so we will close also if started from
                         * the code editor - we really wouldn't be connected to the CalcTool
                         * server if we did not launch from the editor, but just being anal
                         * here.
                         */
                        if (editorLaunch == true)
                        {
                            Application.Exit();
                        }

                        break;
                }
            }
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        /*
         * Button text now reads "Copy Check Box Sets From Clipboard":
         * 
         * To be used when this app is run as a stand-alone application. The user
         * would select Check Box Sets in the code editor and copy that selection to the 
         * clipboard then click this button which will process that code.
         */
        private void copyFromClipboardButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                // Visual feedback on the button they clicked on:
                helpListLabel.Text = string.Format("'{0}' Button Operation", copyFromClipboardButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button operation inspects data copied to the clipboard for check box sets currently being", copyFromClipboardButton.Text));
                helpList.Items.Add(string.Format("controlled. Check box sets will be displayed in the '{0}' list with each set comprising a list element.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("No error messages are generated if this application can not detect check box sets. You simply will not see any box sets inserted into the");
                helpList.Items.Add(string.Format("'{0}' list.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("The '{0}' button gives the user the ability to perform edits on a document's check box control sets", copyFromClipboardButton.Text));
                helpList.Items.Add("which are located at the top of the document's script file in the function getCheckBoxList() as shown in the below example:");
                helpList.Items.Add("");
                helpList.Items.Add("    function getCheckBoxList() {");
                helpList.Items.Add("        return [[\"box15_28\", \"box15_29\"], [\"box15_30\", \"box15_31\"]];");
                helpList.Items.Add("    }");
                helpList.Items.Add("");
                helpList.Items.Add("To edit these sets using the Checkbox Control tool you would first copy the entire function to the clipboard including the ending '}'.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("You would next click the '{0}' button which will insert the above two box sets into the '{1}'", copyFromClipboardButton.Text, setListLabel.Text));
                helpList.Items.Add("list, ready for editing.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("NOTE: One edit you may find useful is to run the '{0}' operation on the '{1}' list to remove check boxes that are no", xmlFilterSetListButton.Text, setListLabel.Text));
                helpList.Items.Add(string.Format("longer on a document - providing that there are boxes in the '{0}' list.", boxListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("If this application was started from an editor:");
                helpList.Items.Add("");
                helpList.Items.Add("    If the current mode is 'Add Check Box Control':");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        The '{0}' button operation will reset a previous '{1}' button click as", copyFromClipboardButton.Text, copySetsButton.Text));
                helpList.Items.Add("        evidenced by the button color changing from the green 'clicked previously' color to the default unclicked color.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        The reasoning is that prior '{0}' editing is not yet complete.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    The current operating mode is not changed.");
                return;
            }
#endregion
            if (Clipboard.ContainsText(TextDataFormat.Text))
            {
                if (copySetsClicked)
                {
                    /*
                     * User is basically undoing a previous 'Copy Check Box Sets To Clipboard'
                     * button click.
                     */
                    copySetsClicked = false;
                    copySetsButton.BackColor = Color.LightGray;
                }

                chkBoxCtrlStr = Clipboard.GetText(TextDataFormat.Text);

                /*
                 * Now inspect the contents of chkBoxCtrlStr for Check Box Sets.
                 * In this case we are not altering the Mode if started from an editor.
                 */
                ProcessInputCode(false);

                if ((launchMode == LaunchMode.disabled) || (launchMode == LaunchMode.addChkBoxCtrl))
                {
                    // Setting focus to the 'State Box' area.
                    stateBox.Select();
                }
            }
        }

        /*
         * This ZIP file event handler fires off from user mouse clicks on the Read-ZIP menu
         * tab. The user selects the appropriate .ZIP file from the Open File Dialog window.
         * The selected .ZIP file will be unzipped and we will obtain any checkboxes from the
         * doc.xml file and insert them into the 'XML Check Boxes' boxList. Once we get the
         * box info this routine will delete the extracted doc.xml file which briefly resides
         * in the temp folder on the C: drive as returned from Path.GetTempPath().
        */
        private void zipFileToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string fileName;

            fileName = GetFileName(curDir, "ZIP");

            if (!String.IsNullOrEmpty(fileName))
            {
                try
                {
                    curDir = Path.GetDirectoryName(fileName);
                    xmlIdLabel.Text = Path.GetFileNameWithoutExtension(fileName).ToUpper();
                    boxList.Items.Clear();
                    boxListFormat = boxFormat.notSet;
                    fileName = UnzipXmlFile(fileName);

                    if (!String.IsNullOrEmpty(fileName))
                    {
                        SetXMLCheckBoxList(fileName);
                        File.Delete(fileName);
                    }

                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }
            }
        }

        /*
         * This XML file event handler fires off from user mouse clicks on the Read-XML menu tab. The user
         * selects the appropriate .XML file from the Open File Dialog window. The selected .XML file will be
         * inspected for checkboxes which will be inserted into the 'XML Check Boxes' boxList.
        */
        private void xMLFilereadCheckboxesToolStripMenuItem_Click(object sender, EventArgs e)
        {
            string fileName;

            fileName = GetFileName(curDir, "XML");

            if (!String.IsNullOrEmpty(fileName))
            {
                try
                {
                    curDir = Path.GetDirectoryName(fileName);
                    xmlIdLabel.Text = Path.GetFileNameWithoutExtension(fileName).ToUpper();
                    boxList.Items.Clear();
                    boxListFormat = boxFormat.notSet;
                    SetXMLCheckBoxList(fileName);
                }
                catch (Exception ex)
                {
                    MessageBox.Show(ex.ToString());
                }
            }
        }

        /*
         * Set the selection mode for boxList. We get the top index of the
         * viewable list, then we clear all selections, then restore the
         * top item as changing the mode somewhow causes the list to shift
         * vertically a couple of items.
        */ 
        private void boxListModeButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - 'Select Single' / 'Select Multiple' Button Operation", boxListLabel.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The 'Select Single/Multiple' button allows you to select boxes in the list for insertion into the '{0}' list.", setManipListLabel.Text));
                helpList.Items.Add("The button text will toggle between 'Select Single' mode and 'Select Multiple' mode each time it is clicked on.");
                helpList.Items.Add("");
                helpList.Items.Add("If the button text is 'Select Single' then box selection is currently in Multiple mode, if the button text is 'Select Multiple' then box");
                helpList.Items.Add("selection is currently in Single mode.");
                helpList.Items.Add("");
                helpList.Items.Add("Each click of the 'Select Single/Multiple' button will first clear any box selections before changing the selection mode.");
                helpList.Items.Add("");
                helpList.Items.Add("Single Box Selection Mode: Each mouse click on a box will either select or un-select that box. You can not select/un-select multiple");
                helpList.Items.Add("    boxes with one click.");
                helpList.Items.Add("");
                helpList.Items.Add("Multiple Box Selection Mode - Using the Shift Key: This will select a contiguous range of boxes in the list. First click on a box, then");
                helpList.Items.Add("    while holding the Shift key down, click on another box either above or below the first box you selected. You will notice now that");
                helpList.Items.Add("    you have a contiguous range of boxes selected. If you click on any box now, you will see that just that box will be selected.");
                helpList.Items.Add("");
                helpList.Items.Add("Multiple Box Selection Mode - While Depressing The Left Mouse Button: This will also select a contiguous range of boxes in the list.");
                helpList.Items.Add("    First position the cursor over any box in the list, then depress the left mouse button to select that box. Keeping the left mouse button");
                helpList.Items.Add("    depressed, move the cursor up or down and you will see the boxes being selected as a contiguous group. If you click on any box");
                helpList.Items.Add("    now, you will see that just that box will be selected.");
                helpList.Items.Add("");
                helpList.Items.Add("Multiple Box Selection Mode - Using the Ctrl Key: While holding the Ctrl key down, click on several boxes and you will see that already");
                helpList.Items.Add("    selected boxes will be un-selected and any un-selected boxes will be selected. Again, to just select one box you only need to click");
                helpList.Items.Add("    on that desired box and any other selected boxes will be un-selected.");
                return;
            }
#endregion
            int topIndex;

            topIndex = boxList.TopIndex;
            boxList.ClearSelected();

            /*
             * Determine the current selectmode configuration, it will be either "MultiExtended" ->
             * multiple items can be selected at a time; or "MultiSimple" -> only one item
             * at a time can be selected.
            */
            switch (boxList.SelectionMode)
            {
                case (SelectionMode.MultiExtended):

                    // Currently in Multiple mode, switching to Single mode.
                    boxList.SelectionMode = SelectionMode.MultiSimple;
                    boxListModeButton.Text = "Select Multiple";
                    break;

                case (SelectionMode.MultiSimple):

                    // Currently in Single mode, switching to Multiple mode.
                    boxList.SelectionMode = SelectionMode.MultiExtended;
                    boxListModeButton.Text = "Select Single";
                    break;
            }

            boxList.TopIndex = topIndex;
        }

        /* 
         * Adds boxes from the 'XML Check Boxes' list to the 'Set Manipulation' list
         * if the boxes are not already in the list. Any pre-existing box selections
         * are maintained after the addition of the XML boxes.
         * 
         * Users are able to select more than one box at a time from the
         * 'XML Check Boxes' list.
        */
        private void setManipXmlAddButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipXmlAddButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button allows you to add selected boxes from the '{1}' list to the '{2}' list.", setManipXmlAddButton.Text, boxListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("The boxes will be added in sorted order.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("You will get a Warning message if a box is already in the '{0}' list.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("Duplicate boxes will not be allowed.");
                return;
            }
#endregion
            // Holds the selected boxes in the 'Set Manipulation' list.
            List<string> selBoxList = new List<string>();

            /*
             * A work list for holding 'Set Manipulation' list boxes
             * and the selected boxes.
             */
            List<string> tmpList = new List<string>();

            if (boxList.SelectedIndices.Count > 0)
            {
                /*
                 * Grab all selected boxes from the 'Set Manipulation' list,
                 * we will use these to reselect them after the box add(s).
                 */
                foreach (string box in setManipList.SelectedItems)
                {
                    selBoxList.Add(box);
                }

                // Grab all boxes in the 'Set Manipulation' list.
                tmpList = setManipList.Items.OfType<string>().ToList();

                /*
                 * Add to tmpList the selected boxes in the 'XML Check Boxes'
                 * list, if they are not in tmpList already.
                 */
                foreach (string box in boxList.SelectedItems.Cast<String>().ToList())
                {
                    if (tmpList.Contains(box))
                    {
                        ProcessErrors("W", setManipListLabel.Text + ": " + setManipXmlAddButton.Text, 2, box, setManipListLabel.Text);
                    }
                    else
                    {
                        tmpList.Add(box);
                    }
                }

                /*
                 * We were sorting this list, but I really do not see
                 * the benefit in doing so especially now that we have 
                 * the new box format. Box sets fully reside on a single
                 * page on a document and the 'XML Check Boxes' list is
                 * sorted by page number which makes sense in that case.
                 */
                //tmpList.Sort(new BoxCompare());

                /*
                 * Clear, then re-populate the 'Set Manipulation' list
                 * and restore any prior selections. Our event handler
                 * for drawing the box items will handle any pre-existing
                 * duplicate box coloring along with handling the selected
                 * boxes.
                 */
                setManipList.Items.Clear();

                for (int i = 0; i < tmpList.Count; i++)
                {
                    setManipList.Items.Add(tmpList[i]);

                    if (selBoxList.Contains(tmpList[i]))
                    {
                        setManipList.SetSelected(i, true);
                    }
                }
            }
        }

        /*
         * Deletes all boxes from the 'Set Manipulation' list.
         */
        private void setManipDelListButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipDelListButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button deletes all boxes from the '{1}' list.", setManipDelListButton.Text,  setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("NOTE: If any boxes in the '{0}' list were duplicate boxes in other '{1}' sets [indicated by the color RED] those", setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add(string.Format("'{0}' sets will be reset to their default color since we are clearing the '{1}' list.", setListLabel.Text, setManipListLabel.Text));
                return;
            }
#endregion
            if (setManipList.Items.Count > 0)
            {
                setManipList.Items.Clear();

                // Clear the global lists dupIndxList and dupManipList.
                dupIndxList.Clear();
                dupManipList.Clear();

                // Refresh the setList to reflect the clearing out of the dupIndxList.
                setList.Refresh();
            }
        }

        // Deletes selected boxes from the 'Set Manipulation' list.
        private void setManipDelSelButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipDelSelButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button deletes all selected boxes from the '{1}' list.", setManipDelSelButton.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("NOTE: If any selected boxes in the '{0}' list were duplicate boxes in other '{1}' sets [indicated by the color RED]", setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add(string.Format("those '{0}' sets will be reset to their default color since we are deleting the '{1}' list boxes that were found to", setListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("be duplicate boxes during a prior '{0}', '{1}', or '{2}' operation.", setManipUpdateButton.Text, setManipAddAboveButton.Text, setManipAddBelowButton.Text));
                return;
            }
#endregion
            if (setManipList.SelectedItems.Count > 0)
            {
                List<string> tmpList = new List<string>();
                List<string> selList = new List<string>();
                bool refreshSetList = false;

                /*
                 * Using this to remember the last setList index we
                 * removed from dupIndxList. This is because the same
                 * setList index may be in dupIndxList for other
                 * dupManipList boxes that belong to the same set.
                 */
                int delIndex = -1;

                // Get all selected boxes.
                foreach (string box in setManipList.SelectedItems)
                {
                    selList.Add(box);
                }

                foreach (string box in setManipList.Items)
                {
                    if (selList.Contains(box))
                    {
                        /* dupManipList holds boxes that are in the 'Set Manipulation' list
                         * that were discovered to be in 'Check Box Sets' other than the currently
                         * selected set while performing the 'Update Box Set' operation. We are 
                         * checking if the box selected for deletion from the 'Set Manipulation'
                         * list is in this list. If it is, we will delete it from the dupManipList
                         * and also remove it's indices from the dupIndxList. It is expected that
                         * there will only be one index, but this will handle multiple instances.
                        */
                        if (dupManipList.Contains(box))
                        {
                            delIndex = -1;
                            dupManipList.Remove(box);

                            for (int i = dupIndxList.Count - 1; i > -1; i--)
                            {
                                if (setList.Items[dupIndxList[i]].ToString().Contains(box))
                                {
                                    if (delIndex != dupIndxList[i])
                                    {
                                        delIndex = dupIndxList[i];
                                        dupIndxList.RemoveAt(i);
                                        refreshSetList = true;
                                    }
                                }
                            }
                        }
                    }
                    else
                    {
                        // Not deleting this box from the list.
                        tmpList.Add(box);
                    }
                }

                // Re-populate the list minus the deletions.
                setManipList.Items.Clear();

                foreach (string box in tmpList)
                {
                    setManipList.Items.Add(box);
                }

                // Debug Start.
                //for (int i = 0; i < dupManipList.Count; i++)
                //{
                //    ProcessErrors("W", "setManipDelSelButton_Click()", 0, string.Format("dupManipList[{0}]: {1}", i, dupManipList[i]), null);
                //}
                //for (int i = 0; i < dupIndxList.Count; i++)
                //{
                //    ProcessErrors("W", "setManipDelSelButton_Click()", 0, string.Format("dupIndxList[{0}]: {1}", i, dupIndxList[i]), null);
                //}
                // Debug Stop.

                // Need to refresh the setList if we deleted duplicate boxes.
                if (refreshSetList)
                {
                    setList.Refresh();
                }
            }
        }

        /* 
         * Updates the currently selected set in the 'Check Box Sets' list
         * with the check boxes in the 'Set Manipulation' list. We only update
         * with boxes that are not already in another box set, else we inform
         * the user about those boxes.
         * 
         * If the 'Check Box Sets' list is currently empty, we will just insert
         * the box set.
        */ 
        private void setManipUpdateButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipUpdateButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS EMPTY - The '{1}' button operation will simply insert the '{2}' list boxes as a", setListLabel.Text, setManipUpdateButton.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("new box set into the '{0}' list subject to the below error checking:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to insert a new set from a '{0}' list that contains only ONE box.", setManipListLabel.Text));
                helpList.Items.Add("    Box sets MUST have more than one box.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS NOT EMPTY - The '{1}' button operation will replace all the boxes in the currently selected", setListLabel.Text, setManipUpdateButton.Text));
                helpList.Items.Add(string.Format("box set in the '{0}' list with the boxes in the '{1}' list subject to the below error checking:", setListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    Each box set in the '{0}' list, EXCEPT for the box set being updated, is inspected for duplicate boxes as compared to the", setListLabel.Text));
                helpList.Items.Add(string.Format("    '{0}' list. An error message is generated if a duplicate box is found and both the duplicate '{1}' box and", setManipListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("    the '{0}' set are colored RED. Duplicate '{1}' boxes are NOT included in the update.", setListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to update a set from a '{0}' list that contains only ONE", setManipListLabel.Text));
                helpList.Items.Add(string.Format("    box - RED duplicate '{0}' boxes are NOT included in the count. Box sets MUST have more than one box.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    NOTE: It will be possible to update a box set even if there are duplicate box errors AS LONG AS there are two or more boxes in");
                helpList.Items.Add(string.Format("    the '{0}' list that are not duplicated in the '{1}' list.", setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("The following is true regardless of the '{0}' list being empty or not empty prior to the update attempt:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied and the update operation will be ABORTED if there was a one box update error.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied after an update of a box set REGARDLESS of any duplicate box errors.", setManipListLabel.Text));
                return;
            }
#endregion
            if (setManipList.Items.Count > 0)
            {
                int selIndex;
                string boxSet;
                List<string> manipBoxList = new List<string>();

                /* 
                 * Index value for currently selected set. Only one set can
                 * be selected at a time in this list. Note if this set is
                 * empty we will default to 0 for the index and just insert
                 * the set as a new set. There ALWAYS is a selected set as long
                 * as there are sets in the list.
                */
                selIndex = (setList.Items.Count > 0 ? setList.SelectedIndex : 0);

                /*
                 * This will return a 'Set Manipulation' box list that has no
                 * duplicate boxes already present in the 'Check Box Sets' list.
                 * The currently selected set is not checked for duplicate boxes.
                 */
                manipBoxList = ProcessDuplicateBoxes(setManipListLabel.Text + ": " + setManipUpdateButton.Text, setListLabel.Text, selIndex);

                if (manipBoxList.Count > 0)
                {
                    if (manipBoxList.Count > 1)
                    {
                        // Remove this index from oneBoxIndxList if it is in there.
                        if (oneBoxIndxList.Contains(selIndex))
                        {
                            oneBoxIndxList.Remove(selIndex);
                        }

                        boxSet = "[" + string.Join(",", manipBoxList.ToArray()) + "]";

                        // Update the selected set/insert new set with the boxes in manipBoxList and select it.
                        if (setList.Items.Count > 0)
                        {
                            setList.Items[selIndex] = boxSet;
                        }
                        else
                        {
                            setList.Items.Add(boxSet);

                            /*
                             * setList WAS empty so set the setListFormat from first
                             * box in manipBoxList:
                             */
                            if (Regex.IsMatch(manipBoxList[0], @"box\d+_\d+"))
                            {
                                setListFormat = boxFormat.oldFormat;
                            }
                            else
                            {
                                setListFormat = boxFormat.newFormat;
                            }
                        }

                        /*
                         * Set the horizontal scrollbar if required:
                         */
                        SetListHorExtent();

                        setList.SetSelected(selIndex, true);
                    }
                    else
                    {
                        ProcessErrors("E", setManipListLabel.Text + ": " + setManipUpdateButton.Text, 4, null, null);
                    }
                }
            }
        }

        /*
         * This adds a new box set to the 'Check Box Sets' list above the currently selected set.
         * This is similar to the 'Update Box Set' operation in that a duplicate box check is
         * performed. Ony non-duplicate boxes will be added to the new set. The new set will be
         * automatically selected after the new set insertion, however the 'Set Manipulation' list
         * will be cleared to make it easier to build a 'Check Box Sets' list which would occur in
         * consecutive box set adds.
        */
        private void setManipAddAboveButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipAddAboveButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS EMPTY - The '{1}' button operation will simply insert the '{2}' list boxes as", setListLabel.Text, setManipAddAboveButton.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("a new box set into the '{0}' list subject to the below error checking:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to insert a new set from a '{0}' list that contains only ONE box.", setManipListLabel.Text));
                helpList.Items.Add("    Box sets MUST have more than one box.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS NOT EMPTY - The '{1}' button operation will insert the '{2}' list boxes as a",setListLabel.Text, setManipAddAboveButton.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("new box set into the '{0}' list immediately ABOVE the currently selected box set subject to the below error checking:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    Each box set in the '{0}' list, INCLUDING the currently selected box set, is inspected for duplicate boxes as compared to", setListLabel.Text));
                helpList.Items.Add(string.Format("    the '{0}' list. An error message is generated if a duplicate box is found and both the duplicate '{0}' box", setManipListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("    and the '{0}' set are colored RED. Duplicate '{1}' boxes are NOT included in the new box set insertion.", setListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to insert a new set from a '{0}' list that contains only ONE box - RED", setManipListLabel.Text));
                helpList.Items.Add(string.Format("    duplicate '{0}' boxes are NOT included in the count. Box sets MUST have more than one box.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    NOTE: It will be possible to insert a new box set even if there are duplicate box errors AS LONG AS there are two or more boxes in");
                helpList.Items.Add(string.Format("    the '{0}' list that are not duplicated in the '{1}' list.",  setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("The following is true regardless of the '{0}' list being empty or not empty prior to the insertion attempt:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will be emptied after an insertion of a new box set IF there were NO duplicate box errors.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied after an insertion of a new box set IF there WERE duplicate box errors.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied and the insertion operation will be ABORTED if there was a one box insertion error.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    IF a box set was inserted into the '{0}' list, it will automatically be selected.", setListLabel.Text));
                return;
            }
#endregion
            if (setManipList.Items.Count > 0)
            {
                int selIndex;
                string boxStr;
                List<string> manipBoxList = new List<string>();

                /* Index value for currently selected set. Only one set can
                 * be selected at a time in this list. Note if this set is
                 * empty we will default to 0 for the index and just insert
                 * the set as a new set. There ALWAYS is a selected set as long
                 * as there are sets in the list.
                */
                selIndex = (setList.Items.Count > 0 ? setList.SelectedIndex : 0);

                /*
                 * This will return a 'Set Manipulation' box list that has no
                 * duplicate boxes already present in the 'Check Box Sets' list.
                 * The currently selected set is also checked for duplicate boxes.
                 */
                manipBoxList = ProcessDuplicateBoxes(setManipListLabel.Text + ": " + setManipAddAboveButton.Text, setListLabel.Text, -1);

                if (manipBoxList.Count > 0)
                {
                    if (manipBoxList.Count > 1)
                    {
                        /*
                         * Okay to insert - Note that we may not be inserting all 'Set Manipulation'
                         * boxes due to duplicates.
                         * Keep the current 'Set Manipulation' list, the dupManipList, and the dupIndxList
                         * if there are duplicate boxes - we still want to display them.
                        */ 
                        if (!(manipBoxList.Count < setManipList.Items.Count))
                        {
                            setManipList.Items.Clear();
                        }

                        boxStr = "[" + string.Join(",", manipBoxList.ToArray()) + "]";

                        if (setList.Items.Count > 0)
                        {
                            setList.SetSelected(selIndex, false);
                            setList.Items.Insert(selIndex, boxStr);
                        }
                        else
                        {
                            setList.Items.Add(boxStr);

                            /*
                             * setList WAS empty so set the setListFormat from first
                             * box in manipBoxList:
                             */
                            if (Regex.IsMatch(manipBoxList[0], @"box\d+_\d+"))
                            {
                                setListFormat = boxFormat.oldFormat;
                            }
                            else
                            {
                                setListFormat = boxFormat.newFormat;
                            }
                        }

                        /*
                         * Set the horizontal scrollbar in setList if
                         * required:
                         */
                        SetListHorExtent();

                        setList.SetSelected(selIndex, true);

                        // We may need to adjust the indices in dupIndxList.
                        for (int i = 0; i < dupIndxList.Count; i++)
                        {
                            if (!(dupIndxList[i] < selIndex))
                            {
                                dupIndxList[i] += 1;
                            }
                        }

                        // We may need to adjust the indices in oneBoxIndxList.
                        for (int i = 0; i < oneBoxIndxList.Count; i++)
                        {
                            if (!(oneBoxIndxList[i] < selIndex))
                            {
                                oneBoxIndxList[i] += 1;
                            }
                        }
                    }
                    else
                    {
                        ProcessErrors("E", setManipListLabel.Text + ": " + setManipAddAboveButton.Text, 4, null, null);
                    }
                }
            }
        }

        /*
         * This adds a new box set to the 'Check Box Sets' list below the currently selected set.
         * This is similar to the 'Update Box Set' operation in that a duplicate box check is
         * performed. Ony non-duplicate boxes will be added to the new set. The new set will be
         * automatically selected after the new set insertion, however the 'Set Manipulation' list
         * will be cleared to make it easier to build a 'Check Box Sets' list which would occur in
         * consecutive box set adds.
        */
        private void setManipAddBelowButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setManipListLabel.Text, setManipAddBelowButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS EMPTY - The '{1}' button operation will simply insert the '{2}' list boxes as", setListLabel.Text, setManipAddBelowButton.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("a new box set into the '{0}' list subject to the below error checking:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to insert a new set from a '{0}' list that contains only ONE box.", setManipListLabel.Text));
                helpList.Items.Add("    Box sets MUST have more than one box.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("WHEN THE '{0}' LIST IS NOT EMPTY - The '{1}' button operation will insert the '{2}' list boxes as a", setListLabel.Text, setManipAddBelowButton.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("new box set into the '{0}' list immediately BELOW the currently selected box set subject to the below error checking:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    Each box set in the '{0}' list, INCLUDING the currently selected box set, is inspected for duplicate boxes as compared to", setListLabel.Text));
                helpList.Items.Add(string.Format("    the '{0}' list. An error message is generated if a duplicate box is found and both the duplicate '{0}' box", setManipListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add(string.Format("    and the '{0}' set are colored RED. Duplicate '{1}' boxes are NOT included in the new box set insertion.", setListLabel.Text, setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The user will get an error message if attempting to insert a new set from a '{0}' list that contains only ONE box - RED", setManipListLabel.Text));
                helpList.Items.Add(string.Format("    duplicate '{0}' boxes are NOT included in the count. Box sets MUST have more than one box.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    NOTE: It will be possible to insert a new box set even if there are duplicate box errors AS LONG AS there are two or more boxes in");
                helpList.Items.Add(string.Format("    the '{0}' list that are not duplicated in the '{1}' list.", setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("The following is true regardless of the '{0}' list being empty or not empty prior to the insertion attempt:", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will be emptied after an insertion of a new box set IF there were NO duplicate box errors.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied after an insertion of a new box set IF there WERE duplicate box errors.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list will NOT be emptied and the insertion operation will be ABORTED if there was a one box insertion error.", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    IF a box set was inserted into the '{0}' list, it will automatically be selected.", setListLabel.Text));
                return;
            }
#endregion
            if (setManipList.Items.Count > 0)
            {
                int selIndex;
                string boxStr;
                List<string> manipBoxList = new List<string>();

                /* Index value for currently selected set. Only one set can
                 * be selected at a time in this list. Note if this set is
                 * empty we will default to 0 for the index and just insert
                 * the set as a new set. There ALWAYS is a selected set as long
                 * as there are sets in the list.
                */
                selIndex = (setList.Items.Count > 0 ? setList.SelectedIndex : 0);

                /*
                 * This will return a 'Set Manipulation' box list that has no
                 * duplicate boxes already present in the 'Check Box Sets' list.
                 * The currently selected set is also checked for duplicate boxes.
                 */
                manipBoxList = ProcessDuplicateBoxes(setManipListLabel.Text + ": " + setManipAddBelowButton.Text, setListLabel.Text, -1);

                if (manipBoxList.Count > 0)
                {
                    if (manipBoxList.Count > 1)
                    {
                        /*
                         * Okay to insert - Note that we may not be inserting all 'Set Manipulation'
                         * boxes due to duplicates.
                         * Keep the current 'Set Manipulation' list, the dupManipList, and the dupIndxList
                         * if there are duplicate boxes - we still want to display them.
                        */
                        if (!(manipBoxList.Count < setManipList.Items.Count))
                        {
                            setManipList.Items.Clear();
                        }

                        boxStr = "[" + string.Join(",", manipBoxList.ToArray()) + "]";

                        if (setList.Items.Count > 0)
                        {
                            setList.SetSelected(selIndex, false);
                            selIndex += 1;
                            setList.Items.Insert(selIndex, boxStr);
                        }
                        else
                        {
                            setList.Items.Add(boxStr);

                            /*
                             * setList WAS empty so set the setListFormat from first
                             * box in manipBoxList:
                             */
                            if (Regex.IsMatch(manipBoxList[0], @"box\d+_\d+"))
                            {
                                setListFormat = boxFormat.oldFormat;
                            }
                            else
                            {
                                setListFormat = boxFormat.newFormat;
                            }
                        }

                        /*
                         * Set the horizontal scrollbar if required:
                         */
                        SetListHorExtent();

                        setList.SetSelected(selIndex, true);

                        // We may need to adjust the indices in dupIndxList.
                        for (int i = 0; i < dupIndxList.Count; i++)
                        {
                            if (!(dupIndxList[i] < selIndex))
                            {
                                dupIndxList[i] += 1;
                            }
                        }

                        // We may need to adjust the indices in oneBoxIndxList.
                        for (int i = 0; i < oneBoxIndxList.Count; i++)
                        {
                            if (!(oneBoxIndxList[i] < selIndex))
                            {
                                oneBoxIndxList[i] += 1;
                            }
                        }
                    }
                    else
                    {
                        ProcessErrors("E", setManipListLabel.Text + ": " + setManipAddBelowButton.Text, 4, null, null);
                    }
                }
            }
        }

        /*
         * Clears the 'Check Box Sets' list and the 'Set Manipulation' list and
         * associated lists that manage one box sets [from an 'XML Filter' operation]
         * and duplicate boxes from a failed 'Update Set' attempt.
         * 
         * If we are receiving checkboxes from CalcHelper we will send a
         * message to CalcHelper to update the current checkbox control
         * status, in this case all checkbox control will be removed.
         */
        private void deleteSetListButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setListLabel.Text, deleteSetListButton.Text);

                helpList.Items.Clear();

                if ((ctClient != null) && (ctClient.boxSelFromCH == true))
                {
                    helpList.Items.Add(string.Format("The '{0}' button operation deletes all box sets from the '{1}' list and removes all Checkbox Control from Calc Helper.", deleteSetListButton.Text, setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("Your checkbox set edits and set additions are mirrored on Calc Helper as a visual aid so you can 'see' what you are doing regarding");
                    helpList.Items.Add("Checkbox Control.");
                    helpList.Items.Add("");
                    helpList.Items.Add("You will still need to make sure all of your changes are saved in your Visual Studio Code file when you are done with Checkbox Control.");
                }
                else
                {
                    helpList.Items.Add(string.Format("The '{0}' button operation deletes all box sets from the '{1}' list and also deletes all boxes from the ", deleteSetListButton.Text, setListLabel.Text));
                    helpList.Items.Add(string.Format("'{0}' list.", setManipListLabel.Text));
                }

                return;
            }
#endregion
            if (setList.Items.Count > 0)
            {
                if ((ctClient != null) && ctClient.IsConnected())
                {
                    if (ctClient.boxSelFromCH == true)
                    {
                        boxSelMsgTypes msgType = boxSelMsgTypes.delAllBoxSets;
                        boxSetsIndex = -1;
                        xmlFilterSetListButton.Enabled = false;

                        ctClient.SendSelectedIndexMsg(string.Format("{0}|{1}", msgType, boxSetsIndex));
                    }
                }

                setList.Items.Clear();
                setList.HorizontalExtent = 0;

                setManipList.Items.Clear();
                dupIndxList.Clear();
                dupManipList.Clear();
                oneBoxIndxList.Clear();

                setListFormat = boxFormat.notSet;
            }
        }

        /*
         * Deletes the 'Check Box Sets' current selection and selects a new
         * 'Check Box Set' and displays the boxes in the new selection in
         * the 'Set Manipulation' list. setList only allows one set at a
         * time to be selected.
         * 
         * If receiving boxes from CalcHelper:
         * 
         * Send Calc Helper a 'delBoxSet' message along with the deleted
         * set's setList index.
         */
        private void deleteSetListSelButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setListLabel.Text, deleteSetListSelButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button operation deletes the currently selected box set from the '{1}' list.", deleteSetListSelButton.Text, setListLabel.Text));
                helpList.Items.Add("");

                if ((ctClient != null) && (ctClient.boxSelFromCH == true))
                {
                    helpList.Items.Add("Calc Helper unselects the associated checkboxes and displays them as uncontrolled checkboxes. There is no automatic selection of a new");
                    helpList.Items.Add(string.Format("checkbox set and the '{0}' button is disabled. From here you could choose to select a new set to edit or you could click or lasso", xmlFilterSetListButton.Text));
                    helpList.Items.Add("uncontrolled checkboxes for a new box set which would be selected and controlled automatically both on Calc Helper and in Checkbox");
                    helpList.Items.Add("Control.");
                }
                else
                {
                    helpList.Items.Add(string.Format("A new box set is automatically selected and displayed in the '{0}' list.", setManipListLabel.Text));
                }
                return;
            }
#endregion
            if (setList.SelectedIndices.Count > 0)
            {
                ProcessSetListSetDeletion("DelSetBtn", setList.SelectedIndex);
            }
        }

        /*
         * 'XML Filter' Functionality:
         * 
         * Runs thru each set of boxes and checks to see if each box in a set is in the 'XML Check Boxes'
         * list, if it is not it removes the box from the set. The set itself is removed if the last box
         * is removed from the set. Single box sets are allowed here, however, we check for that condition
         * when actually generating the method call code. The removal of a box set can result in a new
         * selected set if the current box set is deleted in which case we refresh the 'Set Manipulation'
         * list with the new set contents. We also handle changes regarding duplicate boxes by refreshing
         * global lists dupIndxList and dupManipList.
         * 
         * 'New Set' Functionality:
         * 
         * If receiving boxes from CalcHelper: The button functionality has been converted
         * to 'New Set'. This allows the user to indicate to Calc Helper that they are
         * finished editing the current check box set in the 'Check Box Sets' setList and
         * they want selections and unselections to affect the new set.
         */
        private void xmlFilterSetListButton_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("{0} - '{1}' Button Operation", setListLabel.Text, xmlFilterSetListButton.Text);

                helpList.Items.Clear();

                if ((ctClient != null) && (ctClient.boxSelFromCH == true))
                {
                    helpList.Items.Add(string.Format("The '{0}' button is enabled when a checkbox set in the '{1}' list is selected or when a new checkbox set has been added to", xmlFilterSetListButton.Text, setListLabel.Text));
                    helpList.Items.Add(string.Format("the list. This allows the user to indicate they are finished editing the selected box set and wish to add a new checkbox set to the ", xmlFilterSetListButton.Text, setListLabel.Text));
                    helpList.Items.Add(string.Format("'{0}' list.", setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("The '{0}' button is disabled once it is clicked and the currently selected checkbox set will be unselected. The new checkbox set will then", xmlFilterSetListButton.Text));
                    helpList.Items.Add(string.Format("be added at the 'bottom' of the '{0}' list. ALL new sets are added at the bottom.", setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("The checkbox sets in the '{0}' list will be sorted based on page number when the '{1}' button is", setListLabel.Text, copySetsButton.Text));
                    helpList.Items.Add("clicked.");
                }
                else
                {
                    helpList.Items.Add(string.Format("The '{0}' button operation checks each box set in the '{1}' list against the boxes in the '{2}' list and", xmlFilterSetListButton.Text, setListLabel.Text, boxListLabel.Text));
                    helpList.Items.Add(string.Format("removes any boxes from the box set that are not found in the '{0}' list.", boxListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("It may be possible that the box removal process may result in a box set that contains only one box, in which case the user will see a");
                    helpList.Items.Add("Warning! message as to such and each one box set will be colored RED as a visual identification aid.");
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("The box removal process may also result in one or more box sets being deleted from the '{0}' list. Set deletions MAY result in", setListLabel.Text));
                    helpList.Items.Add(string.Format("the deletion of the currently selected set in which case a NEW box set will be selected and displayed in the '{0}' list.", setManipListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("The removal of boxes from the currently selected box set will also be performed on the '{0}' list as this list displays the", setManipListLabel.Text));
                    helpList.Items.Add("contents of the currently selected set. This is true AS LONG AS the currently selected set is not changed due to sets being removed");
                    helpList.Items.Add("as a result of the filtering operation.");
                }
                return;
            }
#endregion
            if ((ctClient != null) && ctClient.IsConnected() && (ctClient.boxSelFromCH == true))
            {
                // 'New Set' button:

                /*
                 * To indicate a new set the user clicks this button
                 * to remove the selection status to indicate editing 
                 * is done for this set.
                 * 
                 * So if no sets have been selected or there are no sets
                 * then all the user needs to do is simply select/lasso
                 * checkboxes to control and it will automatically
                 * be a new set and will be selected on both tools.
                 */
                if (setList.SelectedIndices.Count > 0)
                {
                    boxSelMsgTypes msgTypes = boxSelMsgTypes.newCBCSelIndex;

                    setList.ClearSelected();
                    boxSetsIndex = -1;
                    xmlFilterSetListButton.Enabled = false;

                    // Inform Calc Helper:
                    ctClient.SendSelectedIndexMsg(string.Format("{0}|{1}", msgTypes, boxSetsIndex));
                }
            }
            else
            {
                // 'XML Filter' button:

                if ((setList.Items.Count > 0) && (boxList.Items.Count > 0))
                {
                    // Holds the boxes that remain after the XML filtering of each box set in the
                    // 'Check Box Sets' list.
                    List<string> filtBoxSetList = new List<string>();

                    // Holds the boxes in the 'Set Manipulation' list.
                    List<string> manipBoxList = new List<string>();

                    // Holds the selected boxes in the 'Set Manipulation' list.
                    List<string> manipSelBoxList = new List<string>();

                    // Used in looking for duplicate boxes between the box list version
                    // of the 'Set Manipulation' list and the boxes in the 'Check Box Sets'.
                    List<string> boxIntersectList = new List<string>();

                    // This holds the filtered sets and will be used to re-populate the 'Check Box Sets'
                    // list when done with filtering.
                    List<string> filtSetsList = new List<string>();

                    // Holds the boxes that will be filtered out as we scan through the box sets
                    // in the 'Check Box Sets' list.
                    List<string> delBoxList = new List<string>();

                    // Temporarily holds boxes from a setList set.
                    List<string> tmpSet = new List<string>();

                    // The index of the currently selected set.
                    int selIndex = setList.SelectedIndex;
                    int newSelIndex = selIndex;

                    // This reflects the indices of the filtered set list since we may be
                    // deleting some sets while filtering.
                    int newIndex = 0;

                    // Indicates if we ended up removing the currently selected set, which is displayed in the
                    // 'Set Manipulation' list unless it was deleted by the user.
                    bool newSelection = false;

                    MatchCollection matches;

                    // We will rebuild this as required.
                    oneBoxIndxList.Clear();

                    for (int i = 0; i < setList.Items.Count; i++)
                    {
                        // Check each box in this set against the 'XML Check Boxes' list:
                        if (setListFormat == boxFormat.newFormat)
                        {
                            matches = Regex.Matches(setList.Items[i].ToString(), @"box\d+");
                        }
                        else
                        {
                            matches = Regex.Matches(setList.Items[i].ToString(), @"box\d+_\d+");
                        }

                        foreach (Match boxMatch in matches)
                        {
                            // MessageBox.Show(String.Format("match.Value: {0}", match.Value));
                            if (boxList.Items.Contains(boxMatch.Value))
                            {
                                // Keeping this box in the set.
                                filtBoxSetList.Add(boxMatch.Value);
                            }
                            else
                            {
                                // Removing this box from the set.
                                delBoxList.Add(boxMatch.Value);
                            }
                        }

                        if (filtBoxSetList.Count > 0)
                        {
                            if (filtBoxSetList.Count < 2)
                            {
                                // Filtered down to a 1 box set.
                                oneBoxIndxList.Add(newIndex);
                            }

                            newIndex += 1;
                            filtSetsList.Add("[" + string.Join(",", filtBoxSetList.ToArray()) + "]");
                            filtBoxSetList.Clear();
                        }
                        else
                        {
                            // A set was deleted:
                            if (i <= selIndex)
                            {
                                if (newSelIndex > 0)
                                {
                                    newSelIndex -= 1;
                                }

                                if (i == selIndex)
                                {
                                    // Deleting the selected set.
                                    newSelection = true;
                                }
                            }
                        }
                    }

                    //  Let user know if they have one or more single box sets after filtering.
                    if (oneBoxIndxList.Count > 0)
                    {
                        ProcessErrors("W", setListLabel.Text + ": " + xmlFilterSetListButton.Text, 1, oneBoxIndxList.Count.ToString(), setListLabel.Text);
                    }

                    // We will rebuild these as required.
                    dupIndxList.Clear();
                    dupManipList.Clear();

                    if (!newSelection)
                    {
                        if (setManipList.Items.Count > 0)
                        {
                            // The list before we process any box deletions due to filtering.
                            manipBoxList = setManipList.Items.OfType<string>().ToList();

                            // This list allows multiple boxes to be selected, so we grab all currently selected boxes.
                            foreach (string box in setManipList.SelectedItems)
                            {
                                manipSelBoxList.Add(box);
                            }

                            // Remove any deleted boxes from manipBoxList. This will now be the list after filtering.
                            manipBoxList = manipBoxList.Except(delBoxList).ToList();

                            // Adjust the selected setManipList list to reflect any deletions.
                            manipSelBoxList = manipSelBoxList.Except(delBoxList).ToList();

                            // For each set we will check if we have a duplicate box EXCEPT for the currently selected set.
                            for (int i = 0; i < filtSetsList.Count; i++)
                            {
                                if (i != newSelIndex)
                                {
                                    tmpSet = filtSetsList[i].Replace("[", "").Replace("]", "").Split(',').ToList();
                                    boxIntersectList = manipBoxList.Intersect(tmpSet).ToList();

                                    if (boxIntersectList.Count > 0)
                                    {
                                        foreach (string box in boxIntersectList)
                                        {
                                            if (!dupManipList.Contains(box))
                                            {
                                                dupManipList.Add(box);
                                            }

                                            dupIndxList.Add(i);
                                        }
                                    }
                                }
                            }

                            // Redraw the setManipList which will highlight any duplicate boxes in the setList.
                            setManipList.Items.Clear();

                            foreach (string box in manipBoxList)
                            {
                                setManipList.Items.Add(box);
                            }

                            // Re-select previously selected boxes that have not been filtered out:
                            foreach (string box in manipSelBoxList)
                            {
                                setManipList.SetSelected(manipBoxList.IndexOf(box), true);
                            }
                        }
                    }

                    /*
                     * Now display the filtered 'Check Box Sets' list.
                     * Doing this down here because the event handler
                     * that draws these items [setList_DrawItem()]
                     * refers to dupIndxList.
                     */
                    setList.Items.Clear();
                    setList.HorizontalExtent = 0;

                    /*
                     * IF the filtSetsList contains sets we will keep
                     * the setList box format as is, else we will reset
                     * it to '.notSet' since it will truly be empty.
                     */
                    if (filtSetsList.Count > 0)
                    {
                        foreach (string set in filtSetsList)
                        {
                            setList.Items.Add(set);
                        }

                        /*
                         * Set the horizontal scrollbar if required:
                         */
                        SetListHorExtent();

                        /*
                         * This may be the same selected set before the filtering
                         * operation even though the index value has changed due
                         * to set deletion(s).
                         */
                        setList.SetSelected(newSelIndex, true);
                    }
                    else
                    {
                        setListFormat = boxFormat.notSet;
                    }

                    if (newSelection)
                    {
                        ProcessSetListSelection();
                    }
                }
            }
        }

        /*
         * This copies the Check Box Control method along with the method invocation
         * to the clipboard. The application will automatically close if started from
         * a code editor AND the 'Copy Check Box Sets To Clipboard' button has been
         * clicked previously - but we first perform this check:
         *
         *      We require a state box;
         *      
         * The user is notified via an error message if the above check failed, the
         * operation is aborted, but we keep the application running if started from a
         * code editor so the situation can be rectumfied.
         */
        private void methButton_Click(object sender, EventArgs e)
        {
            string clipBoardStr = "";
            Match match = null;

#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("'{0}' Button Operation", copyMethCallButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button operation copies both the check box control method code and the code that calls the", copyMethCallButton.Text));
                helpList.Items.Add("method to the clipboard. The code that calls the method directly follows it in the document's script file enabling you to paste these code");
                helpList.Items.Add("segments into your document's Javascript via a single Ctrl+V operation. There is no option to paste these segments individually.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("The method code is stored in a Javascript file, '{0}', located at: '{1}'.", methodFileName, methodPath));
                helpList.Items.Add(string.Format("The state box in the '{0}' entry field is used in the generation of the {1}( ) invocation code.", stateBoxLabel.Text, methodName));
                helpList.Items.Add("");
                helpList.Items.Add("If the application was launched from a code editor it may or may not subsequently close (auto Exit) as explained below:");
                helpList.Items.Add("");
                helpList.Items.Add("    Auto Exiting now depends on what mode the application is operating under after it was launched from an editor.");
                helpList.Items.Add("");
                helpList.Items.Add("    The two editor launch modes are 'Add Check Box Control' and 'Edit Check Box Sets'.");
                helpList.Items.Add("");
                helpList.Items.Add("    The mode will be set to 'Edit Check Box Sets' if the application launched with identifiable check box sets discovered in the clipboard.");
                helpList.Items.Add("    The assumption being that the document already has check box control code and the user wants to edit the pre-existing check box sets");
                helpList.Items.Add("    residing in the function getCheckBoxList().");
                helpList.Items.Add("");
                helpList.Items.Add("    The mode will default to 'Add Check Box Control' if the clipboard was empty or identifiable check box sets were not detected.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' button is only enabled for 'Add Check Box Control' mode.", copyMethCallButton.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    Auto Exit behaviour for 'Add Check Box Control' mode:");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        Auto Exit will occur IF the '{0}' button has already been clicked as indicated by it's green button", copySetsButton.Text));
                helpList.Items.Add("        coloring.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        Auto Exit will NOT occur IF the '{0}' button has not already been clicked as indicated by it's normal", copySetsButton.Text));
                helpList.Items.Add(string.Format("        button coloring. In this case the '{0}' button will be colored green to indicate it was clicked and the", copyMethCallButton.Text));
                helpList.Items.Add(string.Format("        application will stay running, giving the user the ability to click the '{0}' button after which", copySetsButton.Text));
                helpList.Items.Add("        the application will auto Exit.");
                helpList.Items.Add("");
                helpList.Items.Add("    We now use the mode feature because the function getCheckBoxList() is located in one area of the document script file separate from");
                helpList.Items.Add("    where the check box control method and the code that calls it are located.");
                helpList.Items.Add("");
                helpList.Items.Add("    The 'Mode' menu button is disabled if this application is not launched from an editor.");
                helpList.Items.Add("    The user is free to change the mode prior to auto Exiting.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("A successful '{0}' button operation passes the following error checks:", copyMethCallButton.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' entry field must contain a box, 'box1_10', for example.", stateBoxLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("An appropriate error message will be displayed if the above error check failed and the '{0}' button", copySetsButton.Text));
                helpList.Items.Add("operation will be aborted. The application will NOT Exit if started from a code editor to allow for corrections.");
                return;
            }
#endregion
            // First get the method code:
            try
            {
                Clipboard.Clear();
                clipBoardStr = File.ReadAllText(methodPath + methodFileName);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Error reading method file.", MessageBoxButtons.OK, MessageBoxIcon.Error);

                // Either return or exit since we can not provide what the button promises.
                if (editorLaunch)
                {
                    Application.Exit();
                }

                /*
                 * Odd behaviour, we need this here to ALSO return if started from an editor,
                 * [even though we call Application.Exit()] else this code will silently
                 * continue to run and if there are any errors encountered below you will
                 * not see the message(s), if no errors it will still build the invocation
                 * code and stick it in the clipboard.
                */
                return;
            }

            /*
             * Make sure the State Box is in the 'correct' format.
             * Note that we have implemented a box format approach for
             * the 'XML Check Boxes' and the 'Check Box Sets' lists since
             * boxes can populate them from different 'sources'. So we will
             * 'attempt' to check the state box accordingly.
             */
            int formatFlag = 0;

            formatFlag = (boxListFormat == boxFormat.notSet ? 0 : 10) + (setListFormat == boxFormat.notSet ? 0 : 1);

            switch (formatFlag)
            {
                case (0):
                    // Hmmm..no box format info at all:
                    match = Regex.Match(stateBox.Text, @"^\s*(box\d+)\s*$");

                    if (match.Success == false)
                    {
                        match = Regex.Match(stateBox.Text, @"^\s*(box\d+_\d+)\s*$");
                    }
                    
                    break;

                case (1):
                    // Using the setListFormat:
                    if (setListFormat == boxFormat.newFormat)
                    {
                        match = Regex.Match(stateBox.Text, @"^\s*(box\d+)\s*$");
                    }
                    else
                    {
                        match = Regex.Match(stateBox.Text, @"^\s*(box\d+_\d+)\s*$");
                    }

                    break;

                case (10):
                case (11):
                    /*
                     * Using the boxListFormat - We have ensured that both lists
                     * if populated, need to have the same box format, so defaulting
                     * to the boxList.
                     */
                    if (boxListFormat == boxFormat.newFormat)
                    {
                        match = Regex.Match(stateBox.Text, @"^\s*(box\d+)\s*$");
                    }
                    else
                    {
                        match = Regex.Match(stateBox.Text, @"^\s*(box\d+_\d+)\s*$");
                    }

                    break;

            }

            if (match.Success == true)
            {
                /*
                 * State box checked out ok, so append the Method Call code and exit
                 * if started from a code editor AND the user has already clicked
                 * the 'Copy Check Box Sets To Clipboard' button - if not then we
                 * will flag this as clicked and color the button green and let the
                 * user do the check box sets.
                 */
                clipBoardStr += string.Format("\n\n\t\t\t// Before we do anything we make sure the state box still exists on the form,\n\t\t\t// else we do nothing with check box control.\n\t\t\tif ({0} !== 'undefined') ", match.Value);
                clipBoardStr += "{\n\t\t\t\tcheckBoxCtrl(formInstanceId, ";
                clipBoardStr += string.Format("\"{0}\", getCheckBoxList());\n", match.Value);
                clipBoardStr += "\t\t\t}\n";

                if (!String.IsNullOrEmpty(clipBoardStr))
                {
                    Clipboard.SetText(clipBoardStr, TextDataFormat.Text);
                }

                if (editorLaunch)
                {
                    /*
                     * If not started from a code editor the Mode button is disabled, but
                     * here we need to do the following:
                     * 
                     * The 'Copy Method And Method Call To Clipboard' button is not active
                     * if we are in LaunchMode.editChkBoxSets mode, so seeing that we are here
                     * means we need to check if the 'Copy Check Box Sets To Clipboard' button
                     * has been clicked yet, if it has we will auto exit else we will color
                     * this box to give a visual clue that it has been clicked already.
                     */
                    if (copySetsClicked)
                    {
                        Application.Exit();
                    }
                    else
                    {
                        copyMethClicked = true;

                        /*
                         * Give a visual clue that they clicked this button. From here, 
                         * if they click the 'Copy Check Box Sets To Clipboard'
                         * button we will auto exit.
                         */
                        copyMethCallButton.BackColor = Color.GreenYellow;
                    }
                }
            }
            else // Bad state box...
            {
                ProcessErrors("E", copyMethCallButton.Text, 8, stateBox.Text, null);
            }
        }

        /*
         * First copy the 'Check Box Sets' list contents to the clipboard then
         * the application will close if started from the code editor subject to the
         * following [also see the 'In addition' comment below]:
         * 
         *      A) The 'Check Box Sets' list needs to be populated;
         *      B) We make sure there are no one box sets;
         *      C) We check for box set intersections - each set must be distinct from the others.
         *          We stop checking at the first intersection and handle this like a duplicate
         *          box situation.
         *      
         * The user is notified via an error message if any of the above checks fail, the
         * operation is aborted, but we keep the application running if started from the code
         * editor so the situation can be rectumfied.
         *
         * If all checks out and we launched from the editor, we will sort the
         * checkbox sets based on page number. Each set resides on the same page.
         * This 'straightens things out' given that there is nothing stopping
         * the user from inserting checkbox sets from all over the document in
         * a jumbled mess. This will make it much easier the next time somebody
         * works on the sets, especially if receiving checkbox selections from
         * Calc Helper, they can just flip thru the pages from first page to
         * last page.
         * 
         * In addition: If we launched from the code editor we will NOT automatically exit
         * if we are in 'Add Check Box Control' mode AND we did not yet click the
         * 'Copy Method And Call To Clipboard' button.
        */
        private void copySetsButton_Click(object sender, EventArgs e)
        {
            string fcnStr = "";
            List<string> curSet = new List<string>();
            List<string> checkSet = new List<string>();
            List<string> boxIntersectList = new List<string>();
            bool dupFound = false;

            //MessageBox.Show("Entering copySetsButton_Click", "copySetsButton_Click()", MessageBoxButtons.OK, MessageBoxIcon.Information);

#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("'{0}' Button Operation", copySetsButton.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' button operation generates the function getCheckBoxList() which contains the check box lists", copySetsButton.Text));
                helpList.Items.Add("generated by this application. The getCheckBoxList() function is located near the top of the document's Javascript file.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("Below is an example of the getCheckBoxList() function generated from '{0}' '[box1_20,box1_21]' and '[box2_1,box2_3]':", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    function getCheckBoxList() {");
                helpList.Items.Add("        return [[\"box1_20\", \"box1_21\"], [\"box2_1\", \"box2_3\"]];");
                helpList.Items.Add("    }");
                helpList.Items.Add("");
                helpList.Items.Add("The generated code is then copied to the clipboard from where it can be pasted OVER the pre-existing function getCheckBoxList() via");
                helpList.Items.Add("Ctrl+V.");
                helpList.Items.Add("");
                helpList.Items.Add("NOTE: This function is always present in a document's script file even if there are no boxes on the document.");
                helpList.Items.Add("");

                helpList.Items.Add("If the application was launched from a code editor it may or may not subsequently close (auto Exit) as explained below:");
                helpList.Items.Add("");
                helpList.Items.Add("    Auto Exiting now depends on what mode the application is operating under after it was launched from an editor.");
                helpList.Items.Add("");
                helpList.Items.Add("    The two editor launch modes are 'Add Check Box Control' and 'Edit Check Box Sets'.");
                helpList.Items.Add("");
                helpList.Items.Add("    The mode will be set to 'Edit Check Box Sets' if the application launched with identifiable check box sets discovered in the clipboard.");
                helpList.Items.Add("    The assumption being that the document already has check box control code and the user wants to edit the pre-existing check box sets");
                helpList.Items.Add("    residing in the function getCheckBoxList().");
                helpList.Items.Add("");
                helpList.Items.Add("    The mode will default to 'Add Check Box Control' if the clipboard was empty or identifiable check box sets were not detected.");
                helpList.Items.Add("");
                helpList.Items.Add("    The application will auto Exit if it is currently in 'Edit Check Box Sets' mode since there is nothing else to do as the");
                helpList.Items.Add(string.Format("    '{0}' button is disabled in this mode.", copyMethCallButton.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    If the application is currently in 'Add Check Box Control' mode:");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        Auto Exit will occur IF the '{0}' button has already been clicked as indicated by it's green button", copyMethCallButton.Text));
                helpList.Items.Add("        coloring.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("        Auto Exit will NOT occur IF the '{0}' button has not already been clicked as indicated by it's normal", copyMethCallButton.Text));
                helpList.Items.Add(string.Format("        button coloring. In this case the '{0}' button will be colored green to indicate it was clicked and the", copySetsButton.Text));
                helpList.Items.Add(string.Format("        application will stay running, giving the user the ability to click the '{0}' button after which", copyMethCallButton.Text));
                helpList.Items.Add("        the application will auto Exit.");
                helpList.Items.Add("");
                helpList.Items.Add("    We now use the mode feature because the function getCheckBoxList() is located in one area of the document script file separate from");
                helpList.Items.Add("    where the check box control method itself is located. The code that calls the check box control method immediately follows the check");
                helpList.Items.Add(string.Format("    box control method so we now generate these both together - refer to the button Help for the '{0}'", copyMethCallButton.Text));
                helpList.Items.Add("    button.");
                helpList.Items.Add("");
                helpList.Items.Add("    The 'Mode' menu button is disabled if this application is not launched from an editor.");
                helpList.Items.Add("    The user is free to change the mode prior to auto Exiting.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("A successful '{0}' button operation passes the following error checks:", copySetsButton.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    The '{0}' list must contain box sets.", setListLabel.Text));
                helpList.Items.Add(string.Format("    The '{0}' list must not contain 1 box sets.", setListLabel.Text));
                helpList.Items.Add(string.Format("    The '{0}' list must not contain box sets that intersect other box sets.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("An appropriate error message will be displayed if any of the above error checks fail and the '{0}' button", copySetsButton.Text));
                helpList.Items.Add("operation will be aborted. The application will NOT Exit if started from a code editor to allow for corrections.");
                return;
            }
#endregion
            try
            {
                Clipboard.Clear();
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "Error clearing clipboard:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            // A) Make sure we have 'Check Box Sets':
            if (setList.Items.Count > 0)
            {
                // B) Make sure there are no single box sets:
                if (!(oneBoxIndxList.Count > 0))
                {
                    // C) Making sure there are no box set intersections:
                    dupIndxList.Clear();
                    dupManipList.Clear();

                    // We checked the count up above.
                    // For 2 or more sets we will check:
                    for (int i = 0; i < setList.Items.Count - 1; i++)
                    {
                        curSet = setList.Items[i].ToString().Replace("[", "").Replace("]", "").Split(',').ToList();
                        for (int j = i + 1; j < setList.Items.Count; j++)
                        {
                            checkSet = setList.Items[j].ToString().Replace("[", "").Replace("]", "").Split(',').ToList();
                            boxIntersectList = curSet.Intersect(checkSet).ToList();

                            if (boxIntersectList.Count > 0)
                            {
                                foreach (string box in boxIntersectList)
                                {
                                    if (!dupManipList.Contains(box))
                                    {
                                        dupManipList.Add(box);

                                        ProcessErrors("E", copySetsButton.Text, 3, string.Format("{0} is already present in '{1}'", box, setListLabel.Text), null);
                                    }
                                }

                                dupIndxList.Add(j);
                                dupFound = true;
                                setList.SetSelected(i, true);
                                ProcessSetListSelection();
                                setList.Refresh();
                                break;
                            }
                        }

                        if (dupFound)
                        {
                            // Break out of outer loop.
                            break;
                        }
                    }

                    if (!dupFound)
                    {
                        bool doSort = true;
                        string tmpStr = "";

                        /*
                         * All checked out ok, so we will do a page sort on the
                         * checkbox sets if we launched from the editor and then
                         * set the Clipboard with the checkbox sets code and exit
                         * if appropriate to do so - see below.
                         */
                        if (editorLaunch == true)
                        {
                            Dictionary<string, int> boxSetDict = new Dictionary<string, int>();
                            List<string> boxSetList = new List<string>();
                            int setPage;

                            foreach (string boxSet in setList.Items)
                            {
                                boxSetList = boxSet.Replace("[", "").Replace("]", "").Split(',').ToList();

                                /*
                                 * We can use just one box to get the page number.
                                 * When we launched from the editor we built this
                                 * dictionary when we read the document's xml file.
                                 * It contains every checkbox on the document along
                                 * with it's page number. It is the same exact file
                                 * read by Calc Helper.
                                 */
                                if (boxPageDict.TryGetValue(boxSetList[0], out setPage) == true)
                                {
                                    if (!boxSetDict.ContainsKey(boxSet))
                                    {
                                        boxSetDict.Add(boxSet, setPage);
                                    }
                                }
                                else
                                {
                                    // Hmmmmm...
                                    doSort = false;
                                    //MessageBox.Show(string.Format("boxPageDict failed on box: {0}", boxSetList[0]));
                                    break;
                                }
                            }

                            if (doSort == true)
                            {
                                foreach (KeyValuePair<string, int> kvp in boxSetDict.OrderBy(key => key.Value))
                                {
                                    //MessageBox.Show(string.Format("page: {0}\nconcatenating to tmpStr:\n\n{1}", kvp.Value, kvp.Key));
                                    tmpStr += string.Format(", {0}", kvp.Key.Replace(",b", ", b"));
                                }

                                tmpStr = tmpStr.Remove(0, 2);
                            }
                        }
                        else
                        {
                            doSort = false;
                        }

                        if (doSort == false)
                        {
                            tmpStr = string.Join(", ", setList.Items.OfType<string>()).Replace(",b", ", b");
                        }

                        if (setListFormat == boxFormat.newFormat)
                        {
                            tmpStr = Regex.Replace(tmpStr, @"box\d+", delegate(Match match)
                            {
                                return string.Format("\"{0}\"", match.Value);
                            });
                        }
                        else
                        {
                            tmpStr = Regex.Replace(tmpStr, @"box\d+_\d+", delegate(Match match)
                            {
                                return string.Format("\"{0}\"", match.Value);
                            });
                        }

                        fcnStr = "\t\tfunction getCheckBoxList() {\n\t\t\treturn [" + tmpStr + "];\n\t\t}\n";

                        if (!String.IsNullOrEmpty(fcnStr))
                        {
                            // Clipboard.SetText first clears the Clipboard.
                            Clipboard.SetText(fcnStr, TextDataFormat.Text);
                        }

                        if (editorLaunch)
                        {
                            if (launchMode == LaunchMode.editChkBoxSets)
                            {
                                Application.Exit();
                            }
                            else if (launchMode == LaunchMode.addChkBoxCtrl)
                            {
                                if (copyMethClicked)
                                {
                                    Application.Exit();
                                }
                                else
                                {
                                    copySetsClicked = true;

                                    /*
                                     * Give a visual clue that they clicked this button. From here, 
                                     * if they click the 'Copy Method And Method Call To Clipboard'
                                     * button we will auto exit.
                                     */
                                    copySetsButton.BackColor = Color.GreenYellow;
                                }
                            }
                            else
                            {
                                /*
                                 * Should not happen: We will at least give the user the heads up,
                                 * nothing they can do here except tell somebody about it.
                                 */
                                ProcessErrors("E", copySetsButton.Text, 9, null, null);
                            }
                        }
                    }
                }
                else // There are 1 box sets....
                {
                ProcessErrors("E", copySetsButton.Text, 7, setListLabel.Text, oneBoxIndxList.Count.ToString());
                }
            }
            else // There are no 'Check Box Sets' in the list....
            {
                ProcessErrors("E", copySetsButton.Text, 6, null, null);
            }
        }

        /*
         * When NOT receiving box selections from Calc Helper:
         * 
         * Basically display the boxes from the set that was clicked on in the
         * 'Set Manipulation' list.
         * 
         * When receiving box selections from Calc Helper:
         * 
         * We inform Calc Helper of the selection and we enable
         * the 'New Set' button to allow the user to indicate
         * they want to start a new set.
         */
        private void setList_Click(object sender, EventArgs e)
        {
#region help
            /*
             * We do not return after showing help information for a listbox
             * like we do for button help.
             */
            if (showHelp)
            {
                helpListLabel.Text = string.Format("'{0}' List Operation", setListLabel.Text);

                helpList.Items.Clear();

                if ((ctClient != null) && (ctClient.boxSelFromCH == true))
                {
                    helpList.Items.Add(string.Format("The '{0}' list displays the box sets used for CheckBox Control. The box sets come from the following sources:", setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("  From The Clipboard:");
                    helpList.Items.Add("");
                    helpList.Items.Add("    The sets can be retrieved from the clipboard when Checkbox Control is launched from the Visual Studio Code editor. You would first");
                    helpList.Items.Add("    copy the entire Checkbox Control box set list to the clipboard before launching this application, though you may choose to let Calc");
                    helpList.Items.Add(string.Format("    Helper provide the entire '{0}' list contents as discussed below.", setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("  From Calc Helper When You Asked For Checkbox Selections:");
                    helpList.Items.Add("");
                    helpList.Items.Add("    After Calc Helper acknowledges your request for checkbox selections it will retrieve all checkbox sets from your code editor file.");
                    helpList.Items.Add("    The sets will be filtered to remove duplicate and orphan boxes.");
                    helpList.Items.Add("");
                    helpList.Items.Add("  From Checkbox Selections On Calc Helper:");
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("    All you need to do is click or lasso uncontrolled checkboxes on Calc Helper to add checkbox sets to the '{0}' list when a", setListLabel.Text));
                    helpList.Items.Add(string.Format("    checkbox set is not being edited. You can easily remove checkboxes from the '{0}' list by clicking/lassoing controlled", setListLabel.Text));
                    helpList.Items.Add("    checkboxes on Calc Helper when a checkbox set is currently being edited. A Checkbox Controlled checkbox is easy to spot on Calc");
                    helpList.Items.Add(string.Format("    Helper since it is color filled silver. You can tell when a checkbox set is being edited by looking at the '{0}' list or by looking", setListLabel.Text));
                    helpList.Items.Add("    at Calc Helper where you would see the selected checkboxes in the selected set. ");
                    helpList.Items.Add("");
                    helpList.Items.Add("Box selection:");
                    helpList.Items.Add("");
                    helpList.Items.Add("    1) Only ONE box set can be selected at a time.");
                    helpList.Items.Add("    2) The currently selected box set will be selected and displayed on Calc Helper. Calc Helper will display the appropriate document page");
                    helpList.Items.Add("        if the selected set is not on the currently displayed checkbox page.");
                    helpList.Items.Add(string.Format("    3) The current selection will be unselected when you click the '{0}' button.", xmlFilterSetListButton.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("Box set coloring schemes:");
                    helpList.Items.Add("");
                    helpList.Items.Add("    A selected box set will ALWAYS use WHITE for the text [foreground color], else the text will be black for unselected sets.");
                    helpList.Items.Add("");
                    helpList.Items.Add("    Selected or unselected box sets will have a RED background if the following condition is true:");
                    helpList.Items.Add("");
                    helpList.Items.Add("        You click or lasso a single uncontrolled checkbox from Calc Helper.");
                    helpList.Items.Add("");
                    helpList.Items.Add("    A selected box set will have a BLUE background if the above RED background condition does not apply.");
                }
                else
                {
                    helpList.Items.Add(string.Format("The '{0}' list displays the box sets used for CheckBox Control. The box sets can come from the clipboard in one of 2 ways:", setListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("    1) Retrieved from the clipboard via the '{0}' button operation. Click this button while viewing Help for", copyFromClipboardButton.Text));
                    helpList.Items.Add("       details.");
                    helpList.Items.Add("");
                    helpList.Items.Add("    2) Retrieved from the clipboard when Checkbox Control is launched from the Visual Studio Code editor. You would first copy the entire");
                    helpList.Items.Add("       Checkbox Control box set list to the clipboard before launching this application.");
                    helpList.Items.Add("");

                    helpList.Items.Add(string.Format("The '{0}' list can also receive checkbox sets by way of the '{1}' list and it's associated buttons. Select that list and", setListLabel.Text, setManipListLabel.Text));
                    helpList.Items.Add("it's buttons while viewing Help for details.");
                    helpList.Items.Add("");

                    helpList.Items.Add("Box selection:");
                    helpList.Items.Add("");

                    helpList.Items.Add("    1) Only ONE box set can be selected at a time.");
                    helpList.Items.Add("    2) There will always be a selected box set if the list is not empty.");
                    helpList.Items.Add(string.Format("    3) The boxes in the selected box set will be displayed in the '{0}' list for set editing purposes.", setManipListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add("Box set coloring schemes:");
                    helpList.Items.Add("");
                    helpList.Items.Add("    A selected box set will ALWAYS use WHITE for the text [foreground color], else the text will be black for unselected sets.");
                    helpList.Items.Add("");
                    helpList.Items.Add("    Selected or unselected box sets will have a RED background if one of the following conditions is true:");
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("        The box set is a one box set after an '{0}' operation. Click the '{1}' button while viewing Help for details.", xmlFilterSetListButton.Text, xmlFilterSetListButton.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("        The box set may contain a duplicate box as compared to the boxes displayed in the '{0}' list.", setManipListLabel.Text));
                    helpList.Items.Add("");
                    helpList.Items.Add(string.Format("            This would happen after performing an '{0}' / '{1}' / '{2}' button operation.", setManipUpdateButton.Text, setManipAddAboveButton.Text, setManipAddBelowButton.Text));
                    helpList.Items.Add("            Refer to the Button Help for these buttons for details.");
                    helpList.Items.Add("");
                    helpList.Items.Add("    A selected box set will have a BLUE background if the above RED background conditions do not apply.");
                }
            }
#endregion
            if (setList.Items.Count > 0)
            {
                if ((ctClient != null) && ctClient.IsConnected() && (ctClient.boxSelFromCH == true))
                {
                    // Being anal here, just in case...
                    if (setList.SelectedIndices.Count == 1)
                    {
                        boxSelMsgTypes msgType = boxSelMsgTypes.newCBCSelIndex;
                        boxSetsIndex = setList.SelectedIndex;

                        //MessageBox.Show(string.Format("Sending to CH:\n\nboxSelMsgTypes.newCBCSelIndex\n\nboxSetsIndex = {0}", boxSetsIndex));
                        /*
                         * Allow the user to ask for a new set when they
                         * are done editing the currently selected set:
                         */
                        xmlFilterSetListButton.Enabled = true;

                        // Tell Calc Helper we selected a different box set:
                        ctClient.SendSelectedIndexMsg(string.Format("{0}|{1}", msgType, boxSetsIndex));
                    }
                }
                else
                {
                    /*
                     * Now we can clear the dupIndxList and the dupManipList since
                     * we will be redrawing the 'Set Manipulation' list using the
                     * selected set in setList ['Check Box Sets'].
                     */
                    dupIndxList.Clear();
                    dupManipList.Clear();

                    // Redraw the setList to reflect the clearing of the above lists.
                    setList.Refresh();

                    ProcessSetListSelection();
                }
            }
        }

        private void setList_DrawItem(object sender, DrawItemEventArgs e)
        {
            //MessageBox.Show(string.Format("setList_DrawItem - sender = {0}", sender));
            
            try
            {
                // -1 signifies an empty ListBox.
                if ((e.Index > -1) && (e.Index < setList.Items.Count))
                {
                    bool isSelected = ((e.State & DrawItemState.Selected) == DrawItemState.Selected);
                    bool isErrorSet = dupIndxList.Contains(e.Index) || oneBoxIndxList.Contains(e.Index);

                    string text = setList.Items[e.Index].ToString();

                    // Default colors.
                    SolidBrush backgroundBrush = new SolidBrush(Color.White);
                    SolidBrush textBrush = new SolidBrush(Color.Black);

                    // All possible combinations...
                    int colorSetup = (isSelected ? 10 : 0) + (isErrorSet ? 1 : 0);

                    // case (0) is our default colors we set up above.
                    switch (colorSetup)
                    {
                        case (1):
                            // Not selected, an error set.
                            backgroundBrush = new SolidBrush(Color.Red);
                            break;

                        case (10):
                            // Selected, not an error set.
                            backgroundBrush = new SolidBrush(SystemColors.Highlight);
                            textBrush = new SolidBrush(Color.White);
                            break;

                        case (11):
                            // Selected, an error set.
                            backgroundBrush = new SolidBrush(Color.Red);
                            textBrush = new SolidBrush(Color.White);
                            break;
                    }

                    // Draw the background:
                    e.Graphics.FillRectangle(backgroundBrush, e.Bounds);

                    e.Graphics.DrawString(text, e.Font, textBrush, e.Bounds, StringFormat.GenericDefault);

                    // Putting our paints away...clean up the brushes...
                    backgroundBrush.Dispose();
                    textBrush.Dispose();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "setList_DrawItem:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        private void setManipList_DrawItem(object sender, DrawItemEventArgs e)
        {
            // -1 signifies an empty ListBox.
            if (e.Index > -1)
            {
                bool isSelected = ((e.State & DrawItemState.Selected) == DrawItemState.Selected);
                string text = setManipList.Items[e.Index].ToString();
                bool isErrorSet = dupManipList.Contains(text);

                // Default colors.
                SolidBrush backgroundBrush = new SolidBrush(Color.White);
                SolidBrush textBrush = new SolidBrush(Color.Black);

                // All possible combinations...
                int colorSetup = (isSelected ? 10 : 0) + (isErrorSet ? 1 : 0);

                // case (0) is our default colors we set up above.
                switch (colorSetup)
                {
                    case (1):
                        // Not selected, an error set.
                        backgroundBrush = new SolidBrush(Color.Red);
                        break;

                    case (10):
                        // Selected, not an error set.
                        backgroundBrush = new SolidBrush(SystemColors.Highlight);
                        textBrush = new SolidBrush(Color.White);
                        break;

                    case (11):
                        // Selected, an error set.
                        backgroundBrush = new SolidBrush(Color.Red);
                        textBrush = new SolidBrush(Color.White);
                        break;
                }

                // Draw the background:
                e.Graphics.FillRectangle(backgroundBrush, e.Bounds);

                e.Graphics.DrawString(text, e.Font, textBrush, e.Bounds, StringFormat.GenericDefault);

                // Putting our paints away...clean up the brushes...
                backgroundBrush.Dispose();
                textBrush.Dispose();
            }
        }

        // I added this just to show the help info.
        private void setManipList_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("'{0}' List Operation", setManipListLabel.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' list displays the boxes for a CheckBox Control set. This list allows you to update a selected '{1}' set or to", setManipListLabel.Text, setListLabel.Text));
                helpList.Items.Add(string.Format("add a new set to the '{0}' list.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("Boxes are inserted into the '{0}' list under the following conditions:", setManipListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    1) When a '{0}' set is selected - all boxes in the set will be displayed in the list.", setListLabel.Text));
                helpList.Items.Add(string.Format("    2) When the '{0}' list '{1}' button is clicked - all selected boxes in the '{2}' list will be added to the list.", setManipListLabel.Text, setManipXmlAddButton.Text, boxListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("Box selection - Multiple boxes can be selected at a time:");
                helpList.Items.Add("");
                helpList.Items.Add("    Using the Shift Key:");
                helpList.Items.Add("");
                helpList.Items.Add("        This will select a contiguous range of boxes in the list. First click on a box, then while holding the Shift key down, click on another");
                helpList.Items.Add("        box either above or below the first box you selected. You will notice now that you have a contiguous range of boxes selected.");
                helpList.Items.Add("        If you click on any box now, you will see that just that box will be selected.");
                helpList.Items.Add("");
                helpList.Items.Add("    While Depressing The Left Mouse Button:");
                helpList.Items.Add("");
                helpList.Items.Add("        This will also select a contiguous range of boxes in the list. First position the cursor over any box in the list, then depress the left");
                helpList.Items.Add("        mouse button to select that box. Keeping the left mouse button depressed, move the cursor up or down and you will see the boxes");
                helpList.Items.Add("        being selected as a contiguous group. If you click on any box now, you will see that just that box will be selected.");
                helpList.Items.Add("");
                helpList.Items.Add("    Using the Ctrl Key:");
                helpList.Items.Add("");
                helpList.Items.Add("        While holding the Ctrl key down, click on several boxes and you will see that already selected boxes will be un-selected and any");
                helpList.Items.Add("        un-selected boxes will be selected. Again, to just select one box you only need to click on that desired box and any other selected");
                helpList.Items.Add("        boxes will be un-selected.");
                helpList.Items.Add("");
                helpList.Items.Add("Box set coloring schemes:");
                helpList.Items.Add("");
                helpList.Items.Add("    A selected box will ALWAYS use WHITE for the text [foreground color], else the text will be black for unselected boxes.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    A selected box will have a BLUE background IF that box is not duplicated [see below] in the '{0}' list.", setListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("    A box will have a RED background if the box is duplicated in the '{0}' list. This would happen after performing an", setListLabel.Text));
                helpList.Items.Add(string.Format("    '{0}' / '{1}' / '{2}' button operation. Refer to the Button Help for these buttons for details.", setManipUpdateButton.Text, setManipAddAboveButton.Text, setManipAddBelowButton.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    The RED background will apply whether the box is selected or unselectd.");
            }
#endregion
        }

        // I added this just to show the help info.
        private void boxList_Click(object sender, EventArgs e)
        {
#region help
            if (showHelp)
            {
                helpListLabel.Text = string.Format("'{0}' List Operation", boxListLabel.Text);

                helpList.Items.Clear();

                helpList.Items.Add(string.Format("The '{0}' list displays ALL check boxes found in a document's XML file.", boxListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("The document name [minus the .XML] will be displayed in the label area just above the boxes. A document that has NO check boxes will");
                helpList.Items.Add("result in an empty list with just the name displayed.");
                helpList.Items.Add("");
                helpList.Items.Add(string.Format("Boxes are inserted into the '{0}' list under the following conditions:", boxListLabel.Text));
                helpList.Items.Add("");
                helpList.Items.Add("    1) When the CheckBox Control application is started from the SciTE editor:");
                helpList.Items.Add("");
                helpList.Items.Add("        The application first looks for the document's .XML file in the folder where the Ruby script file being edited is located, which will");
                helpList.Items.Add("        usually be C:\\temp\\Work but not a guarantee. If ABCD1002.RBW is being editied, it will look for ABCD1002.XML.");
                helpList.Items.Add("");
                helpList.Items.Add("        If the above search failed, the application will search the C:\\temp\\Work folder looking for the document's .XML file.");
                helpList.Items.Add("");
                helpList.Items.Add("        If the .XML was still not found the search will be repeated for the .ZIP file, searching the same folders as above in the same order.");
                helpList.Items.Add("        If ABCD1002.RBW is being editied, it will look for ABCD1002.ZIP.");
                helpList.Items.Add("");
                helpList.Items.Add("        If the .ZIP file was found the application will unzip the file and attempt to extract the doc.XML file and extract all check boxes within.");
                helpList.Items.Add("");
                helpList.Items.Add("    2) Using the 'Read Zip File / XML File' menu options:");
                helpList.Items.Add("");
                helpList.Items.Add("        The 'Zip File' option will automatically extract the .XML file for you.");
                helpList.Items.Add(string.Format("        Either option will search the .XML file for check boxes and insert them into the '{0}' list with the document name minus", boxListLabel.Text));
                helpList.Items.Add("        the file extension displayed in the label area.");
                helpList.Items.Add("");
                helpList.Items.Add("Box Selection:");
                helpList.Items.Add("");
                helpList.Items.Add("        Please click the 'Select Single / Select Multiple' button [beneath the box list] while viewing Help for box selection information.");
            }
#endregion
        }

        private void addCheckBoxControlToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*
             * Mode button is disabled by design if launched as a stand-alone app.
             * Successful editor launching will enable it.
             */
            //if (launchMode == LaunchMode.editChkBoxSets)
            //{
            //    launchMode = LaunchMode.addChkBoxCtrl;
            //    ProcessMode(launchMode);
            //}
            ProcessMode(LaunchMode.addChkBoxCtrl, false);
        }

        private void editCheckBoxSetsToolStripMenuItem_Click(object sender, EventArgs e)
        {
            /*
             * Mode button is disabled by design if launched as a stand-alone app.
             * Successful editor launching will enable it.
             */
            //if (launchMode == LaunchMode.addChkBoxCtrl)
            //{
            //    launchMode = LaunchMode.editChkBoxSets;
            //    ProcessMode(launchMode);
            //}
            ProcessMode(LaunchMode.editChkBoxSets, false);
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            /*
             * First send current check box sets to Calc Helper
             * if we are receiving checkbox selections so Calc
             * Helper can set it's check box state list. Just a
             * safety thing in case the user doesn't actually
             * copy the Checkbox Control sets to the code file.
             * This would cause some confusion when they refreshed 
             * Calc Helper from a visual viewpoint.
             * 
             * When Calc Helper is sending box selections to
             * Checkbox Control, the user is seeing their
             * checkbox additions and deletions in 'real
             * time' on the document displayed in Calc Helper.
             * But Calc Helper's checkbox state list is at the last
             * refresh just before it started sending checkbox
             * selections to Checkbox Control. IF the user doesn't
             * update the code editor file which holds the REAL
             * checkbox sets, a refresh will not 'look'
             * right, for example deleted checkboxes [from being
             * controlled] would look like they are not controlled
             * but Calc Helper would think they were because I do
             * a list comparison - the checkboxes read in from the
             * code file set list is compared against the state list
             * and any adds/deletes are dealt with.
             */
            if (ctClient != null)
            {
                if (ctClient.boxSelFromCH == true)
                {
                    SendSetStateMsg();
                }

                ctClient.StopClient();
            }
        }

        /*
         * Ask CalcHelper for checkbox selections
         * or tell Calc Helper we do not want any more selections.
         */
        private void boxSelFromCHBtn_Click(object sender, EventArgs e)
        {
            if (showHelp == true)
            {
                // Preventing unexpected activity.....
                return;
            }

            if ((ctClient != null) && ctClient.IsConnected())
            {
                if (ctClient.boxSelFromCH == false)
                {
                    ctClient.BoxSelRequest();
                }
                else
                {
                    /*
                     * Send current setList checkbox sets to Calc Helper
                     * so it can update it's checkbox control state list.
                     * This is just in case the user does not update the
                     * real checkbox control sets in the visual studio
                     * code file. Want to keep the 'visual' checkbox control
                     * appearance on Calc Helper consistent with the real
                     * checkbox control when a refresh operation is performed
                     * on Calc Helper.
                     */
                    SendSetStateMsg();

                    ProcessCHBoxSelOnOff(false);

                    ctClient.BoxSelRequestOff();
                    ctClient.boxSelFromCH = false;

                    boxSelFromCHBtn.BackgroundImage = ChkBoxCtrl.Properties.Resources.BoxSelFromCHOff;
                    boxSelFromCHBtn.ToolTipText = "Click To Receive Checkbox Selections From CalcHelper";
                }
            }
        }
    }
}
